#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <limits.h>

#define SECTOR_SIZE 512
#define NUM_SECTORS 1440
#define DISK_SIZE SECTOR_SIZE*NUM_SECTORS

struct bpb {
	u_int16_t	jump;
	char		manname[8];
	u_int16_t	bytes_per_sector;
	u_int8_t	sectors_per_cluster;
	u_int16_t	reserved_sectors;
	u_int8_t	num_fats;
	u_int16_t	max_root_entries;
	u_int16_t	logical_sectors;
	u_int8_t	medium_desc;
	u_int16_t	sectors_per_fat;
	u_int16_t	sectors_per_track;
	u_int16_t	heads;
	u_int16_t	hidden_sectors;
} bpb;

unsigned char *disk_image;
u_int8_t bootsec[SECTOR_SIZE];

void set_default_bpb()
{
	bpb.jump = 0x00;
	strncpy(bpb.manname, "c1opd", 8);
	bpb.bytes_per_sector = SECTOR_SIZE;
	bpb.sectors_per_cluster = 2;
	bpb.reserved_sectors = 1;
	bpb.num_fats = 2;
	bpb.max_root_entries = 112;
	bpb.logical_sectors = NUM_SECTORS;
	bpb.medium_desc = 0xf9;
	bpb.sectors_per_fat = 3;
	bpb.sectors_per_track = 9;
	bpb.heads = 2;
	bpb.hidden_sectors = 0;
}

void clear_bootsec()
{
	int i;

	for (i = 0; i < SECTOR_SIZE; i++)
		bootsec[i] = 0x0;
}

int main(int argc, char *argv[])
{
	int i;
	int is_c1, is_dos;
	int fd;
	char in_filename[PATH_MAX + 1], out_filename[PATH_MAX + 1];
	char *c1_ext = "-c1";
	char *dos_ext = "-c1";
	char *out_ext = "";
	
	if ((disk_image = (unsigned char *) calloc(512*1440, sizeof(unsigned char))) == NULL)
	{
		fprintf(stderr, "%s: Allocating memory for disk image: %s\n", strerror(errno));
		exit(errno);
	}
	
	set_default_bpb();

	clear_bootsec();

	is_c1 = (strcmp(argv[0], "c1todos") == 0) ? 1 : 0;
	is_dos = (strcmp(argv[0], "dostoc1") == 0) ? 1 : 0;

	if ((is_c1 == 0) && (is_dos == 0))
	{
		fprintf(stderr, "%s: Can't determine what to do.\nExecutable must be named either c1todos or dostoc1.\n", argv[0]);
		exit(-1);
	}

	if(is_dos)
	{
		strcpy(bootsec, "C1/ADDER Disk written on the ICL OPD.\n\r");
	}

	if(is_c1)
	{
		bootsec[0x0] = bpb.jump % 256;
		bootsec[0x1] = bpb.jump / 256;
		strncpy(&bootsec[0x3], bpb.manname, 8);
		bootsec[0x0b] = bpb.bytes_per_sector % 256;
		bootsec[0x0c] = bpb.bytes_per_sector / 256;
		bootsec[0x0d] = bpb.sectors_per_cluster;
		bootsec[0x0e] = bpb.reserved_sectors % 256;
		bootsec[0x0f] = bpb.reserved_sectors / 256;
		bootsec[0x10] = bpb.num_fats;
		bootsec[0x11] = bpb.max_root_entries % 256;
		bootsec[0x12] = bpb.max_root_entries / 256;
		bootsec[0x13] = bpb.logical_sectors % 256;
		bootsec[0x14] = bpb.logical_sectors / 256;
		bootsec[0x15] = bpb.medium_desc;
		bootsec[0x16] = bpb.sectors_per_fat % 256;
		bootsec[0x17] = bpb.sectors_per_fat / 256;
		bootsec[0x18] = bpb.sectors_per_track % 256;
		bootsec[0x19] = bpb.sectors_per_track / 256;
		bootsec[0x1a] = bpb.heads % 256;
		bootsec[0x1b] = bpb.heads / 256;
		bootsec[0x1c] = bpb.hidden_sectors % 256;
		bootsec[0x1d] = bpb.hidden_sectors / 256;
	}

	if ((argc == 1) || (argc > 3))
	{
		fprintf(stderr, "%s: Usage: %s <input disk image> [<output disk image>]\n");
		exit(-1);
	}

	strncpy(in_filename, argv[1], PATH_MAX);

	if (argc == 3)
	{
		strncpy(out_filename, argv[2], PATH_MAX);
	}
	else
	{
		if (is_c1)
			out_ext = dos_ext;

		if (is_dos)
			out_ext = c1_ext;
			
		sprintf(out_filename, "%s%s", in_filename, out_ext);
	}

	if ((fd = open(in_filename, O_RDONLY)) < 0)
	{
		fprintf(stderr, "%s: Unable to open input disk image %s: %s\n", argv[0], in_filename, strerror(errno));
		exit(-1);
	}

	if ((disk_image = (unsigned char *) malloc(DISK_SIZE)) == NULL)
	{
		fprintf(stderr, "%s: unable to allocate memory for disk image: %s\n", argv[0], strerror(errno));
		exit(-1);
	}

	if (read(fd, disk_image, DISK_SIZE) != DISK_SIZE)
	{
		fprintf(stderr, "%s: input disk image is too small or short read: %s\n", argv[0], strerror(errno));
		exit(-1);
	}

	close(fd);

	memcpy(disk_image, bootsec, SECTOR_SIZE);

	if ((fd = creat(out_filename, 0644)) < 0)
	{
		fprintf(stderr, "%s: Unable to create output disk image %s: %s\n", argv[0], out_filename, strerror(errno));
		exit(-1);
	}

	if ((write(fd, disk_image, DISK_SIZE)) != DISK_SIZE)
	{
		fprintf(stderr, "%s: output disk image write failed: %s\n", argv[0], strerror(errno));
	}

	close(fd);

	exit(0);
}
